/* aliases2cdbm.c: Parse an /etc/aliases-type file into cdbmake format.
 * created 1999-Feb-11 01:49 jmk
 * autodate: 1999-Feb-13 04:56
 * 
 * by Jim Knoble <jmknoble@pobox.com>
 * Copyright  1999 Jim Knoble
 * 
 * Disclaimer:
 * 
 * The software is provided "as is", without warranty of any kind,
 * express or implied, including but not limited to the warranties of
 * merchantability, fitness for a particular purpose and
 * noninfringement. In no event shall the author(s) be liable for any
 * claim, damages or other liability, whether in an action of
 * contract, tort or otherwise, arising from, out of or in connection
 * with the software or the use or other dealings in the software.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dynlist.h"

static char *prog_name;

static char *key = NULL;
static char *val = NULL;
static int key_buf_length = 0;
static int val_buf_length = 0;
static int key_index = 0;
static int val_index = 0;

static char version_info[] =
"aliases2cdbm version %s\n"
;

static char usage_info[] =
"\n"
"  Usage: %s [<filename> [<filename>...]]\n"
"         %s {-h|--help}\n"
"         %s {-V|--version}\n"
"\n"
"  Input:\n"
"      One or more text files in `/etc/aliases' syntax.\n"
"\n"
"  Output:\n"
"      Text suitable for use as input to `cdbmake' for creating a CDB\n"
"      database.\n"
"\n"
"  Arguments:\n"
"      A filename of `-' will cause input to be read from stdin.\n"
"\n"
"      If no filename is specified on the command line, input is read\n"
"      from stdin.\n"
"\n"
"      Output is printed on stdout, and warning messages are printed on\n"
"      stderr.\n"
"\n"
;

void version(void)
{
   printf(version_info, VERSION);
}

void usage(void)
{
   printf(usage_info, prog_name, prog_name, prog_name);
   exit(1);
}

int is_eof(int c)
{
   int status = (EOF == c);
   return(status);
}

int is_white(int c)
{
   int status = (isspace(c) && ('\n' != c));
   return(status);
}

int is_continuation_line(int c)
{
   int status = is_white(c);
   return(status);
}

int is_comment_line(int c)
{
   int status = ('#' == c);
   return(status);
}

int is_blank_line(int c)
{
   int status = ('\n' == c);
   return(status);
}

int skip_white(FILE *infile)
{
   int c;
   
   while (EOF != (c = getc(infile)))
    {
       if (!is_white(c))
	{
	   ungetc(c, infile);
	   break;
	}
    }
   return(c);
}

int skip_line(FILE *infile)
{
   int c;
   
   while (EOF != (c = getc(infile)))
    {
       if ('\n' == c)
	{
	   break;
	}
    }
   return(c);
}

int trim_trailing_space(char **s, int *next_index, int truncate)
{
   int i = *next_index - 1;
   
   if ((NULL != *s) && (0 < i))
    {
       while ((0 < i) && isspace((*s)[i - 1]))
	{
	   i--;
	}
       *next_index = i + 1;
       if (truncate)
	{
	   (*s)[i] = '\0';
	}
       else if (isspace((*s)[i]))
	{
	   (*next_index)--;
	}
    }
   return(i);
}

void out_of_memory(void)
{
   fprintf(stderr, "%s: Aaaaahhh!!  I ran out of memory!\n", prog_name);
   exit(1);
}

void warn_eof(char *filename, int line_count)
{
   fprintf(stderr, "%s:%d: warning: no newline at end of file\n",
	   filename, line_count);
}

void error_premature_eol(char *filename, int line_count)
{
   fprintf(stderr, "%s:%d: error: premature end of line\n",
	   filename, line_count);
}

void error_empty_key(char *filename, int line_count)
{
   fprintf(stderr, "%s:%d: error: empty key\n", filename, line_count);
}

void error_empty_value(char *filename, int line_count, char *key)
{
   fprintf(stderr, "%s:%d: error: empty value for key `%s'\n",
	   filename, line_count, key);
}

void error_whitespace_bol(char *filename, int line_count)
{
   fprintf(stderr, "%s:%d: error: whitespace at beginning of line\n",
	   filename, line_count);
}

void print_alias(FILE *outfile, char *key, int key_len, char *val, int val_len)
{
   fprintf(outfile, "+%d,%d:%s->%s\n", key_len, val_len, key, val);
}

void print_end_of_output(FILE *outfile)
{
   fprintf(outfile, "\n");
}

int parse_input(FILE *infile, char *filename, FILE *outfile) 
{
   int c;
   int c0;
   int in_key = 0;
   int in_val = 0;
   int new_line = 0;
   int line_count = 1;
   int error_count = 0;
   int status;
   int key_len = 0;
   int val_len = 0;
   int eof = 0;

   while (!eof)
    {
       c = getc(infile);
       if (in_key)
	{
	   eof = is_eof(c);
	   if (eof)
	    {
	       error_count++;
	       error_premature_eol(filename, line_count);
	       warn_eof(filename, line_count);
	       in_key    = in_val    = 0;
	       key_index = val_index = 0;
	       break;
	    }
	   else if ('\n' == c)
	    {
	       error_count++;
	       error_premature_eol(filename, line_count);
	       line_count++;
	       in_key    = in_val    = 0;
	       key_index = val_index = 0;
	       continue;
	    }
	   if (is_white(c))
	    {
	       c = '\0';
	       in_key = 0;
	       c0 = skip_white(infile);
	       if (':' == c0)
		{
		   getc(infile);
		}
	    }
	   else if (':' == c)
	    {
	       c = '\0';
	       in_key = 0;
	    }
	   status = append_to_buf(&key, &key_buf_length, &key_index, c);
	   if (APPEND_FAILURE == status)
	    {
	       out_of_memory();
	    }
	   if (!in_key)
	    {
	       key_len = key_index - 1;
	       if ((NULL == key) || (0 >= key_len))
		{
		   error_count++;
		   error_empty_key(filename, line_count);
		   skip_line(infile);
		   line_count++;
		   continue;
		}
	       in_val = 1;
	       val_index = 0;
	       skip_white(infile);
	    }
	}
       else if (in_val)
	{
	   eof = is_eof(c);
	   if (eof && !new_line)
	    {
	       warn_eof(filename, line_count);
	       c = '\0';
	       in_val = 0;
	    }
	   else if (new_line) 
	    {
	       if (is_comment_line(c))
		{
		   c0 = skip_line(infile);
		   eof = is_eof(c0);
		   if (eof)
		    {
		       warn_eof(filename, line_count);
		       c = '\0';
		       in_val = 0;
		       new_line = 0;
		    }
		   else 
		    {
		       line_count++;
		       /* new_line remains set */
		       continue;
		    }
		}
	       else if (is_continuation_line(c))
		{
		   c0 = skip_white(infile);
		   if (is_comment_line(c0))
		    {
		       /* new_line remains set */
		       continue;
		    }
		   else
		    {
		       new_line = 0;
		       eof = is_eof(c0);
		       if (eof)
			{
			   warn_eof(filename, line_count);
			}
		       if (eof || is_blank_line(c0))
			{
			   c = '\0';
			   in_val = 0;
			}
		       else
			{
			   c = ' ';
			}
		    }
		}
	       else
		{
		   if (!eof)
		    {
		       ungetc(c, infile);
		    }
		   c = '\0';
		   in_val = 0;
		   new_line = 0;
		}
	    }
	   else if ('\n' == c)
	    {
	       trim_trailing_space(&val, &val_index, 0);
	       line_count++;
	       new_line = 1;
	       continue;
	    }
	   status = append_to_buf(&val, &val_buf_length, &val_index, c);
	   if (APPEND_FAILURE == status)
	    {
	       out_of_memory();
	    }
	   if (!in_val)
	    {
	       val_len = trim_trailing_space(&val, &val_index, 1);
	       if ((NULL == val) || (0 >= val_len))
		{
		   error_count++;
		   error_empty_value(filename, (line_count - 1), key);
		   continue;
		}
	       print_alias(outfile, key, key_len, val, val_len);
	    }
	}
       else
	{
	   eof = is_eof(c);
	   if (eof)
	    {
	       break;
	    }
	   else if (is_continuation_line(c))
	    {
	       c0 = skip_white(infile);
	       if (!(is_blank_line(c0) || is_comment_line(c0)))
		{
		   error_whitespace_bol(filename, line_count);
		}
	       skip_line(infile);
	       line_count++;
	    }
	   else if (is_comment_line(c))
	    {
	       skip_line(infile);
	       line_count++;
	    }
	   else if (is_blank_line(c))
	    {
	       line_count++;
	    }
	   else
	    {
	       in_key = 1;
	       key_index = 0;
	       ungetc(c, infile);
	    }
	}
    }
   return(error_count);
}

int process_input_file(char *filename, FILE *outfile)
{
   FILE *infile;
   char *munged_filename = filename;
   int using_stdin = 0;
   int error_count;

   if (0 == strcmp(filename, "-"))
    {
       using_stdin = 1;
       infile = stdin;
       munged_filename = "stdin";
    }
   else
    {
       infile = fopen(filename, "r");
       if (NULL == infile)
	{
	   perror(filename);
	   return(-1);
	}
    }
   error_count = parse_input(infile, munged_filename, outfile);
   if (!using_stdin)
    {
       fclose(infile);
    }
   return(error_count);
}

void update_status(int error_count, int *status)
{
   if (error_count)
    {
       *status = 1;
    }
}

int process_args(int argc, char **argv)
{
   int i;
   
   for (i = 1; i < argc; i++)
    {
       if ('-' != argv[i][0])
	{
	   break;
	}
       else if ((0 == strcmp(argv[i], "-h")) ||
		(0 == strcmp(argv[i], "-help")) ||
		(0 == strcmp(argv[i], "--help")))
	{
	   version();
	   usage();
	}
       else if ((0 == strcmp(argv[i], "-V")) ||
		(0 == strcmp(argv[i], "-version")) ||
		(0 == strcmp(argv[i], "--version")))
	{
	   version();
	   exit(1);
	}
       else if (0 == strcmp(argv[i], "-"))
	{
	   /* '-' is a filename meaning stdin;
	    * leave it in the list of files to process.
	    */
	   break;
	}
       else if (0 == strcmp(argv[i], "--"))
	{
	   /* End of options. */
	   i++;
	   break;
	}
       else
	{
	   printf("%s: Unknown option `%s'\n", prog_name, argv[i]);
	   usage();
	}
    }
   return(i);
}

int main(int argc, char **argv)
{
   FILE *outfile = stdout;
   int error_count = 0;
   int i;
   int status;

   prog_name = argv[0];
   i = process_args(argc, argv);
   
   if (i == argc)
    {
       /* No files specified on command line; use stdin. */
       error_count = process_input_file("-", outfile);
       update_status(error_count, &status);
    }
   else
    {
       for (i = 1; i < argc; i++)
	{
	   error_count = process_input_file(argv[i], outfile);
	   update_status(error_count, &status);
	}
    }
   print_end_of_output(outfile);
   
   return(status);
}

